# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import freezegun
from unittest import mock

import oslo_messaging as om

from watcher.common import exception
from watcher.common import rpc
from watcher import notifications
from watcher import objects
from watcher.tests.db import base
from watcher.tests.objects import utils


@freezegun.freeze_time('2016-10-18T09:52:05.219414')
class TestActionPlanNotification(base.DbTestCase):

    def setUp(self):
        super(TestActionPlanNotification, self).setUp()
        p_get_notifier = mock.patch.object(rpc, 'get_notifier')
        m_get_notifier = p_get_notifier.start()
        self.addCleanup(p_get_notifier.stop)
        self.m_notifier = mock.Mock(spec=om.Notifier)

        def fake_get_notifier(publisher_id):
            self.m_notifier.publisher_id = publisher_id
            return self.m_notifier

        m_get_notifier.side_effect = fake_get_notifier
        self.goal = utils.create_test_goal(mock.Mock())
        self.audit = utils.create_test_audit(mock.Mock(), interval=None)
        self.strategy = utils.create_test_strategy(mock.Mock())

    def test_send_invalid_action_plan(self):
        action_plan = utils.get_test_action_plan(
            mock.Mock(), state='DOESNOTMATTER', audit_id=1)

        self.assertRaises(
            exception.InvalidActionPlan,
            notifications.action_plan.send_update,
            mock.MagicMock(), action_plan, host='node0')

    def test_send_action_plan_update(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.ONGOING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit, strategy=self.strategy)
        notifications.action_plan.send_update(
            mock.MagicMock(), action_plan, host='node0',
            old_state=objects.action_plan.State.PENDING)

        # The 1st notification is because we created the object.
        # The 2nd notification is because we created the action plan object.
        self.assertEqual(3, self.m_notifier.info.call_count)
        notification = self.m_notifier.info.call_args[1]
        payload = notification['payload']

        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "watcher_object.namespace": "watcher",
                "watcher_object.version": "1.1",
                "watcher_object.data": {
                    "global_efficacy": [],
                    "strategy_uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                    "strategy": {
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.0",
                        "watcher_object.data": {
                            "updated_at": None,
                            "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                            "name": "TEST",
                            "parameters_spec": {},
                            "created_at": "2016-10-18T09:52:05Z",
                            "display_name": "test strategy",
                            "deleted_at": None
                        },
                        "watcher_object.name": "StrategyPayload"
                    },
                    "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061",
                    "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                    "audit": {
                        "watcher_object.data": {
                            "interval": None,
                            "next_run_time": None,
                            "auto_trigger": False,
                            "parameters": {},
                            "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                            "name": "My Audit",
                            "strategy_uuid": None,
                            "goal_uuid": (
                                "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                            "deleted_at": None,
                            "scope": [],
                            "state": "PENDING",
                            "updated_at": None,
                            "created_at": "2016-10-18T09:52:05Z",
                            "audit_type": "ONESHOT"
                        },
                        "watcher_object.name": "TerseAuditPayload",
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.2"
                    },
                    "deleted_at": None,
                    "state": "ONGOING",
                    "updated_at": None,
                    "created_at": "2016-10-18T09:52:05Z",
                    "state_update": {
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.0",
                        "watcher_object.data": {
                            "old_state": "PENDING",
                            "state": "ONGOING"
                        },
                        "watcher_object.name": "ActionPlanStateUpdatePayload"
                    },
                },
                "watcher_object.name": "ActionPlanUpdatePayload"
            },
            payload
        )

    def test_send_action_plan_create(self):
        action_plan = utils.get_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.PENDING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit.as_dict(), strategy=self.strategy.as_dict())
        notifications.action_plan.send_create(
            mock.MagicMock(), action_plan, host='node0')

        self.assertEqual(2, self.m_notifier.info.call_count)
        notification = self.m_notifier.info.call_args[1]
        payload = notification['payload']

        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "watcher_object.namespace": "watcher",
                "watcher_object.version": "1.1",
                "watcher_object.data": {
                    "global_efficacy": [],
                    "strategy_uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                    "strategy": {
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.0",
                        "watcher_object.data": {
                            "updated_at": None,
                            "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                            "name": "TEST",
                            "parameters_spec": {},
                            "created_at": "2016-10-18T09:52:05Z",
                            "display_name": "test strategy",
                            "deleted_at": None
                        },
                        "watcher_object.name": "StrategyPayload"
                    },
                    "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061",
                    "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                    "audit": {
                        "watcher_object.data": {
                            "interval": None,
                            "next_run_time": None,
                            "auto_trigger": False,
                            "parameters": {},
                            "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                            "name": "My Audit",
                            "strategy_uuid": None,
                            "goal_uuid": (
                                "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                            "deleted_at": None,
                            "scope": [],
                            "state": "PENDING",
                            "updated_at": None,
                            "created_at": "2016-10-18T09:52:05Z",
                            "audit_type": "ONESHOT"
                        },
                        "watcher_object.name": "TerseAuditPayload",
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.2"
                    },
                    "deleted_at": None,
                    "state": "PENDING",
                    "updated_at": None,
                    "created_at": None,
                },
                "watcher_object.name": "ActionPlanCreatePayload"
            },
            payload
        )

    def test_send_action_plan_delete(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.DELETED,
            audit_id=self.audit.id, strategy_id=self.strategy.id)
        notifications.action_plan.send_delete(
            mock.MagicMock(), action_plan, host='node0')

        # The 1st notification is because we created the audit object.
        # The 2nd notification is because we created the action plan object.
        self.assertEqual(3, self.m_notifier.info.call_count)
        notification = self.m_notifier.info.call_args[1]
        payload = notification['payload']

        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "watcher_object.namespace": "watcher",
                "watcher_object.version": "1.1",
                "watcher_object.data": {
                    "global_efficacy": [],
                    "strategy_uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                    "strategy": {
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.0",
                        "watcher_object.data": {
                            "updated_at": None,
                            "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3",
                            "name": "TEST",
                            "parameters_spec": {},
                            "created_at": "2016-10-18T09:52:05Z",
                            "display_name": "test strategy",
                            "deleted_at": None
                        },
                        "watcher_object.name": "StrategyPayload"
                    },
                    "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061",
                    "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                    "audit": {
                        "watcher_object.data": {
                            "interval": None,
                            "next_run_time": None,
                            "auto_trigger": False,
                            "parameters": {},
                            "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                            "name": "My Audit",
                            "strategy_uuid": None,
                            "goal_uuid": (
                                "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                            "deleted_at": None,
                            "scope": [],
                            "state": "PENDING",
                            "updated_at": None,
                            "created_at": "2016-10-18T09:52:05Z",
                            "audit_type": "ONESHOT"
                        },
                        "watcher_object.name": "TerseAuditPayload",
                        "watcher_object.namespace": "watcher",
                        "watcher_object.version": "1.2"
                    },
                    "deleted_at": None,
                    "state": "DELETED",
                    "updated_at": None,
                    "created_at": "2016-10-18T09:52:05Z",
                },
                "watcher_object.name": "ActionPlanDeletePayload"
            },
            payload
        )

    def test_send_action_plan_action(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.ONGOING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit, strategy=self.strategy)
        notifications.action_plan.send_action_notification(
            mock.MagicMock(), action_plan, host='node0',
            action='execution', phase='start')

        # The 1st notification is because we created the audit object.
        # The 2nd notification is because we created the action plan object.
        self.assertEqual(3, self.m_notifier.info.call_count)
        notification = self.m_notifier.info.call_args[1]

        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "event_type": "action_plan.execution.start",
                "payload": {
                    "watcher_object.data": {
                        "created_at": "2016-10-18T09:52:05Z",
                        "deleted_at": None,
                        "fault": None,
                        "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                        "audit": {
                            "watcher_object.namespace": "watcher",
                            "watcher_object.name": "TerseAuditPayload",
                            "watcher_object.version": "1.2",
                            "watcher_object.data": {
                                "interval": None,
                                "next_run_time": None,
                                "auto_trigger": False,
                                "parameters": {},
                                "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                                "name": "My Audit",
                                "strategy_uuid": None,
                                "goal_uuid": (
                                    "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                                "deleted_at": None,
                                "scope": [],
                                "state": "PENDING",
                                "updated_at": None,
                                "created_at": "2016-10-18T09:52:05Z",
                                "audit_type": "ONESHOT"
                            }
                        },
                        "global_efficacy": [],
                        "state": "ONGOING",
                        "strategy_uuid": (
                            "cb3d0b58-4415-4d90-b75b-1e96878730e3"),
                        "strategy": {
                            "watcher_object.data": {
                                "created_at": "2016-10-18T09:52:05Z",
                                "deleted_at": None,
                                "display_name": "test strategy",
                                "name": "TEST",
                                "parameters_spec": {},
                                "updated_at": None,
                                "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3"
                            },
                            "watcher_object.name": "StrategyPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "updated_at": None,
                        "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061"
                    },
                    "watcher_object.name": "ActionPlanActionPayload",
                    "watcher_object.namespace": "watcher",
                    "watcher_object.version": "1.1"
                }
            },
            notification
        )

    def test_send_action_plan_action_with_error(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.ONGOING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit, strategy=self.strategy)

        try:
            # This is to load the exception in sys.exc_info()
            raise exception.WatcherException("TEST")
        except exception.WatcherException:
            notifications.action_plan.send_action_notification(
                mock.MagicMock(), action_plan, host='node0',
                action='execution', priority='error', phase='error')

        self.assertEqual(1, self.m_notifier.error.call_count)
        notification = self.m_notifier.error.call_args[1]
        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "event_type": "action_plan.execution.error",
                "payload": {
                    "watcher_object.data": {
                        "created_at": "2016-10-18T09:52:05Z",
                        "deleted_at": None,
                        "fault": {
                            "watcher_object.data": {
                                "exception": "WatcherException",
                                "exception_message": "TEST",
                                "function_name": (
                                    "test_send_action_plan_action_with_error"),
                                "module_name": "watcher.tests.notifications."
                                               "test_action_plan_notification"
                            },
                            "watcher_object.name": "ExceptionPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                        "audit": {
                            "watcher_object.data": {
                                "interval": None,
                                "next_run_time": None,
                                "auto_trigger": False,
                                "parameters": {},
                                "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                                "name": "My Audit",
                                "strategy_uuid": None,
                                "goal_uuid": (
                                    "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                                "deleted_at": None,
                                "scope": [],
                                "state": "PENDING",
                                "updated_at": None,
                                "created_at": "2016-10-18T09:52:05Z",
                                "audit_type": "ONESHOT"
                            },
                            "watcher_object.name": "TerseAuditPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.2"
                        },
                        "global_efficacy": [],
                        "state": "ONGOING",
                        "strategy_uuid": (
                            "cb3d0b58-4415-4d90-b75b-1e96878730e3"),
                        "strategy": {
                            "watcher_object.data": {
                                "created_at": "2016-10-18T09:52:05Z",
                                "deleted_at": None,
                                "display_name": "test strategy",
                                "name": "TEST",
                                "parameters_spec": {},
                                "updated_at": None,
                                "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3"
                            },
                            "watcher_object.name": "StrategyPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "updated_at": None,
                        "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061"
                    },
                    "watcher_object.name": "ActionPlanActionPayload",
                    "watcher_object.namespace": "watcher",
                    "watcher_object.version": "1.1"
                }
            },
            notification
        )

    def test_send_action_plan_cancel(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.ONGOING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit, strategy=self.strategy)
        notifications.action_plan.send_cancel_notification(
            mock.MagicMock(), action_plan, host='node0',
            action='cancel', phase='start')

        # The 1st notification is because we created the audit object.
        # The 2nd notification is because we created the action plan object.
        self.assertEqual(3, self.m_notifier.info.call_count)
        notification = self.m_notifier.info.call_args[1]
        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "event_type": "action_plan.cancel.start",
                "payload": {
                    "watcher_object.data": {
                        "created_at": "2016-10-18T09:52:05Z",
                        "deleted_at": None,
                        "fault": None,
                        "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                        "audit": {
                            "watcher_object.namespace": "watcher",
                            "watcher_object.name": "TerseAuditPayload",
                            "watcher_object.version": "1.2",
                            "watcher_object.data": {
                                "interval": None,
                                "next_run_time": None,
                                "auto_trigger": False,
                                "parameters": {},
                                "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                                'name': 'My Audit',
                                "strategy_uuid": None,
                                "goal_uuid": (
                                    "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                                "deleted_at": None,
                                "scope": [],
                                "state": "PENDING",
                                "updated_at": None,
                                "created_at": "2016-10-18T09:52:05Z",
                                "audit_type": "ONESHOT"
                            }
                        },
                        "global_efficacy": [],
                        "state": "ONGOING",
                        "strategy_uuid": (
                            "cb3d0b58-4415-4d90-b75b-1e96878730e3"),
                        "strategy": {
                            "watcher_object.data": {
                                "created_at": "2016-10-18T09:52:05Z",
                                "deleted_at": None,
                                "display_name": "test strategy",
                                "name": "TEST",
                                "parameters_spec": {},
                                "updated_at": None,
                                "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3"
                            },
                            "watcher_object.name": "StrategyPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "updated_at": None,
                        "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061"
                    },
                    "watcher_object.name": "ActionPlanCancelPayload",
                    "watcher_object.namespace": "watcher",
                    "watcher_object.version": "1.1"
                }
            },
            notification
        )

    def test_send_action_plan_cancel_with_error(self):
        action_plan = utils.create_test_action_plan(
            mock.Mock(), state=objects.action_plan.State.ONGOING,
            audit_id=self.audit.id, strategy_id=self.strategy.id,
            audit=self.audit, strategy=self.strategy)

        try:
            # This is to load the exception in sys.exc_info()
            raise exception.WatcherException("TEST")
        except exception.WatcherException:
            notifications.action_plan.send_cancel_notification(
                mock.MagicMock(), action_plan, host='node0',
                action='cancel', priority='error', phase='error')

        self.assertEqual(1, self.m_notifier.error.call_count)
        notification = self.m_notifier.error.call_args[1]
        self.assertEqual("infra-optim:node0", self.m_notifier.publisher_id)
        self.assertDictEqual(
            {
                "event_type": "action_plan.cancel.error",
                "payload": {
                    "watcher_object.data": {
                        "created_at": "2016-10-18T09:52:05Z",
                        "deleted_at": None,
                        "fault": {
                            "watcher_object.data": {
                                "exception": "WatcherException",
                                "exception_message": "TEST",
                                "function_name": (
                                    "test_send_action_plan_cancel_with_error"),
                                "module_name": "watcher.tests.notifications."
                                               "test_action_plan_notification"
                            },
                            "watcher_object.name": "ExceptionPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "audit_uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                        "audit": {
                            "watcher_object.data": {
                                "interval": None,
                                "next_run_time": None,
                                "auto_trigger": False,
                                "parameters": {},
                                "uuid": "10a47dd1-4874-4298-91cf-eff046dbdb8d",
                                'name': 'My Audit',
                                "strategy_uuid": None,
                                "goal_uuid": (
                                    "f7ad87ae-4298-91cf-93a0-f35a852e3652"),
                                "deleted_at": None,
                                "scope": [],
                                "state": "PENDING",
                                "updated_at": None,
                                "created_at": "2016-10-18T09:52:05Z",
                                "audit_type": "ONESHOT"
                            },
                            "watcher_object.name": "TerseAuditPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.2"
                        },
                        "global_efficacy": [],
                        "state": "ONGOING",
                        "strategy_uuid": (
                            "cb3d0b58-4415-4d90-b75b-1e96878730e3"),
                        "strategy": {
                            "watcher_object.data": {
                                "created_at": "2016-10-18T09:52:05Z",
                                "deleted_at": None,
                                "display_name": "test strategy",
                                "name": "TEST",
                                "parameters_spec": {},
                                "updated_at": None,
                                "uuid": "cb3d0b58-4415-4d90-b75b-1e96878730e3"
                            },
                            "watcher_object.name": "StrategyPayload",
                            "watcher_object.namespace": "watcher",
                            "watcher_object.version": "1.0"
                        },
                        "updated_at": None,
                        "uuid": "76be87bd-3422-43f9-93a0-e85a577e3061"
                    },
                    "watcher_object.name": "ActionPlanCancelPayload",
                    "watcher_object.namespace": "watcher",
                    "watcher_object.version": "1.1"
                }
            },
            notification
        )



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import oslo_i18n
from oslo_i18n import _lazy

# The domain is the name of the App which is used to generate the folder
# containing the translation files (i.e. the .pot file and the various locales)
DOMAIN = "watcher"

_translators = oslo_i18n.TranslatorFactory(domain=DOMAIN)

# The primary translation function using the well-known name "_"
_ = _translators.primary

# The contextual translation function using the name "_C"
_C = _translators.contextual_form

# The plural translation function using the name "_P"
_P = _translators.plural_form


def lazy_translation_enabled():
    return _lazy.USE_LAZY


def get_available_languages():
    return oslo_i18n.get_available_languages(DOMAIN)



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.common import service_manager
from watcher.decision_engine.model.notification import cinder as cnotification
from watcher.decision_engine.model.notification import nova as novanotification
from watcher.tests.decision_engine.model import faker_cluster_state


class FakeManager(service_manager.ServiceManager):

    API_VERSION = '1.0'

    fake_cdmc = faker_cluster_state.FakerModelCollector()

    @property
    def service_name(self):
        return 'watcher-fake'

    @property
    def api_version(self):
        return self.API_VERSION

    @property
    def publisher_id(self):
        return 'test_publisher_id'

    @property
    def conductor_topic(self):
        return 'test_conductor_topic'

    @property
    def notification_topics(self):
        return ['nova']

    @property
    def conductor_endpoints(self):
        return []  # Disable audit endpoint

    @property
    def notification_endpoints(self):
        return [
            novanotification.VersionedNotification(self.fake_cdmc),
        ]


class FakeStorageManager(FakeManager):

    fake_cdmc = faker_cluster_state.FakerStorageModelCollector()

    @property
    def notification_endpoints(self):
        return [
            cnotification.CapacityNotificationEndpoint(self.fake_cdmc),
            cnotification.VolumeCreateEnd(self.fake_cdmc),
            cnotification.VolumeUpdateEnd(self.fake_cdmc),
            cnotification.VolumeDeleteEnd(self.fake_cdmc),
            cnotification.VolumeAttachEnd(self.fake_cdmc),
            cnotification.VolumeDetachEnd(self.fake_cdmc),
            cnotification.VolumeResizeEnd(self.fake_cdmc),
        ]



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Run storage database migration.
"""

import sys

from oslo_config import cfg

from watcher.common import service
from watcher import conf
from watcher.db import migration
from watcher.db import purge

CONF = conf.CONF


class DBCommand(object):

    @staticmethod
    def upgrade():
        migration.upgrade(CONF.command.revision)

    @staticmethod
    def downgrade():
        migration.downgrade(CONF.command.revision)

    @staticmethod
    def revision():
        migration.revision(CONF.command.message, CONF.command.autogenerate)

    @staticmethod
    def stamp():
        migration.stamp(CONF.command.revision)

    @staticmethod
    def version():
        print(migration.version())

    @staticmethod
    def create_schema():
        migration.create_schema()

    @staticmethod
    def purge():
        purge.purge(CONF.command.age_in_days, CONF.command.max_number,
                    CONF.command.goal, CONF.command.exclude_orphans,
                    CONF.command.dry_run)


def add_command_parsers(subparsers):
    parser = subparsers.add_parser(
        'upgrade',
        help="Upgrade the database schema to the latest version. "
             "Optionally, use --revision to specify an alembic revision "
             "string to upgrade to.")
    parser.set_defaults(func=DBCommand.upgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser(
        'downgrade',
        help="Downgrade the database schema to the oldest revision. "
             "While optional, one should generally use --revision to "
             "specify the alembic revision string to downgrade to.")
    parser.set_defaults(func=DBCommand.downgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser('stamp')
    parser.add_argument('revision', nargs='?')
    parser.set_defaults(func=DBCommand.stamp)

    parser = subparsers.add_parser(
        'revision',
        help="Create a new alembic revision. "
             "Use --message to set the message string.")
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.set_defaults(func=DBCommand.revision)

    parser = subparsers.add_parser(
        'version',
        help="Print the current version information and exit.")
    parser.set_defaults(func=DBCommand.version)

    parser = subparsers.add_parser(
        'create_schema',
        help="Create the database schema.")
    parser.set_defaults(func=DBCommand.create_schema)

    parser = subparsers.add_parser(
        'purge',
        help="Purge the database.")
    parser.add_argument('-d', '--age-in-days',
                        help="Number of days since deletion (from today) "
                             "to exclude from the purge. If None, everything "
                             "will be purged.",
                        type=int, default=None, nargs='?')
    parser.add_argument('-n', '--max-number',
                        help="Max number of objects expected to be deleted. "
                             "Prevents the deletion if exceeded. No limit if "
                             "set to None.",
                        type=int, default=None, nargs='?')
    parser.add_argument('-t', '--goal',
                        help="UUID or name of the goal to purge.",
                        type=str, default=None, nargs='?')
    parser.add_argument('-e', '--exclude-orphans', action='store_true',
                        help="Flag to indicate whether or not you want to "
                             "exclude orphans from deletion (default: False).",
                        default=False)
    parser.add_argument('--dry-run', action='store_true',
                        help="Flag to indicate whether or not you want to "
                             "perform a dry run (no deletion).",
                        default=False)

    parser.set_defaults(func=DBCommand.purge)


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help='Available commands',
                                handler=add_command_parsers)


def register_sub_command_opts():
    cfg.CONF.register_cli_opt(command_opt)


def main():
    register_sub_command_opts()
    # this is hack to work with previous usage of watcher-dbsync
    # pls change it to watcher-dbsync upgrade
    valid_commands = set([
        'upgrade', 'downgrade', 'revision',
        'version', 'stamp', 'create_schema',
        'purge',
    ])
    if not set(sys.argv).intersection(valid_commands):
        sys.argv.append('upgrade')

    service.prepare_service(sys.argv, CONF)
    CONF.command.func()



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Watcher test utilities."""

from oslo_utils import timeutils

from watcher.db import api as db_api
from watcher.db.sqlalchemy import models
from watcher import objects


def id_generator():
    id_ = 1
    while True:
        yield id_
        id_ += 1


def _load_relationships(model, db_data):
    rel_data = {}
    relationships = db_api.get_instance()._get_relationships(model)
    for name, relationship in relationships.items():
        related_model = relationship.argument
        if not db_data.get(name):
            rel_data[name] = None
        else:
            rel_data[name] = related_model(**db_data.get(name))

    return rel_data


def get_test_audit_template(**kwargs):
    audit_template_data = {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', 'e74c40e0-d825-11e2-a28f-0800200c9a66'),
        'goal_id': kwargs.get('goal_id', 1),
        'strategy_id': kwargs.get('strategy_id', None),
        'name': kwargs.get('name', 'My Audit Template'),
        'description': kwargs.get('description', 'Desc. Of My Audit Template'),
        'scope': kwargs.get('scope', []),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }

    # ObjectField doesn't allow None nor dict, so if we want to simulate a
    # non-eager object loading, the field should not be referenced at all.
    audit_template_data.update(
        _load_relationships(models.AuditTemplate, kwargs))

    return audit_template_data


def create_test_audit_template(**kwargs):
    """Create test audit template entry in DB and return AuditTemplate DB object.

    Function to be used to create test AuditTemplate objects in the database.
    :param kwargs: kwargsargs with overriding values for audit template's
                   attributes.
    :returns: Test AuditTemplate DB object.
    """
    audit_template = get_test_audit_template(**kwargs)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kwargs:
        del audit_template['id']
    dbapi = db_api.get_instance()
    return dbapi.create_audit_template(audit_template)


def get_test_audit(**kwargs):
    audit_data = {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', '10a47dd1-4874-4298-91cf-eff046dbdb8d'),
        'name': kwargs.get('name', 'My Audit'),
        'audit_type': kwargs.get('audit_type', 'ONESHOT'),
        'state': kwargs.get('state', objects.audit.State.PENDING),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
        'parameters': kwargs.get('parameters', {}),
        'interval': kwargs.get('interval', '3600'),
        'goal_id': kwargs.get('goal_id', 1),
        'strategy_id': kwargs.get('strategy_id', None),
        'scope': kwargs.get('scope', []),
        'auto_trigger': kwargs.get('auto_trigger', False),
        'next_run_time': kwargs.get('next_run_time'),
        'hostname': kwargs.get('hostname', 'host_1'),
        'start_time': kwargs.get('start_time'),
        'end_time': kwargs.get('end_time'),
        'force': kwargs.get('force', False)

    }
    # ObjectField doesn't allow None nor dict, so if we want to simulate a
    # non-eager object loading, the field should not be referenced at all.
    audit_data.update(_load_relationships(models.Audit, kwargs))

    return audit_data


def create_test_audit(**kwargs):
    """Create test audit entry in DB and return Audit DB object.

    Function to be used to create test Audit objects in the database.
    :param kwargs: kwargsargs with overriding values for audit's attributes.
    :returns: Test Audit DB object.
    """
    audit = get_test_audit(**kwargs)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kwargs:
        del audit['id']
    dbapi = db_api.get_instance()
    return dbapi.create_audit(audit)


def get_test_action(**kwargs):
    action_data = {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', '10a47dd1-4874-4298-91cf-eff046dbdb8d'),
        'action_plan_id': kwargs.get('action_plan_id', 1),
        'action_type': kwargs.get('action_type', 'nop'),
        'input_parameters':
            kwargs.get('input_parameters',
                       {'key1': 'val1',
                        'key2': 'val2',
                        'resource_id':
                        '10a47dd1-4874-4298-91cf-eff046dbdb8d'}),
        'state': kwargs.get('state', objects.action_plan.State.PENDING),
        'parents': kwargs.get('parents', []),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }

    # ObjectField doesn't allow None nor dict, so if we want to simulate a
    # non-eager object loading, the field should not be referenced at all.
    action_data.update(_load_relationships(models.Action, kwargs))

    return action_data


def create_test_action(**kwargs):
    """Create test action entry in DB and return Action DB object.

    Function to be used to create test Action objects in the database.
    :param kwargs: kwargsargs with overriding values for action's attributes.
    :returns: Test Action DB object.
    """
    action = get_test_action(**kwargs)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kwargs:
        del action['id']
    dbapi = db_api.get_instance()
    return dbapi.create_action(action)


def get_test_action_plan(**kwargs):
    action_plan_data = {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', '76be87bd-3422-43f9-93a0-e85a577e3061'),
        'state': kwargs.get('state', objects.action_plan.State.ONGOING),
        'audit_id': kwargs.get('audit_id', 1),
        'strategy_id': kwargs.get('strategy_id', 1),
        'global_efficacy': kwargs.get('global_efficacy', []),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
        'hostname': kwargs.get('hostname', 'host_1'),
    }

    # ObjectField doesn't allow None nor dict, so if we want to simulate a
    # non-eager object loading, the field should not be referenced at all.
    action_plan_data.update(_load_relationships(models.ActionPlan, kwargs))

    return action_plan_data


def create_test_action_plan(**kwargs):
    """Create test action plan entry in DB and return Action Plan DB object.

    Function to be used to create test Action objects in the database.
    :param kwargs: kwargsargs with overriding values for action's attributes.
    :returns: Test Action DB object.
    """
    action = get_test_action_plan(**kwargs)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kwargs:
        del action['id']
    dbapi = db_api.get_instance()
    return dbapi.create_action_plan(action)


def get_test_goal(**kwargs):
    return {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', 'f7ad87ae-4298-91cf-93a0-f35a852e3652'),
        'name': kwargs.get('name', 'TEST'),
        'display_name': kwargs.get('display_name', 'test goal'),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
        'efficacy_specification': kwargs.get('efficacy_specification', []),
    }


def create_test_goal(**kwargs):
    """Create test goal entry in DB and return Goal DB object.

    Function to be used to create test Goal objects in the database.
    :param kwargs: kwargs which override default goal values of its attributes.
    :returns: Test Goal DB object.
    """
    goal = get_test_goal(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.create_goal(goal)


def get_test_scoring_engine(**kwargs):
    return {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', 'e8370ede-4f39-11e6-9ffa-08002722cb21'),
        'name': kwargs.get('name', 'test-se-01'),
        'description': kwargs.get('description', 'test scoring engine 01'),
        'metainfo': kwargs.get('metainfo', 'test_attr=test_val'),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }


def create_test_scoring_engine(**kwargs):
    """Create test scoring engine in DB and return ScoringEngine DB object.

    Function to be used to create test ScoringEngine objects in the database.
    :param kwargs: kwargs with overriding values for SE'sattributes.
    :returns: Test ScoringEngine DB object.
    """
    scoring_engine = get_test_scoring_engine(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.create_scoring_engine(scoring_engine)


def get_test_strategy(**kwargs):
    strategy_data = {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', 'cb3d0b58-4415-4d90-b75b-1e96878730e3'),
        'name': kwargs.get('name', 'TEST'),
        'display_name': kwargs.get('display_name', 'test strategy'),
        'goal_id': kwargs.get('goal_id', 1),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
        'parameters_spec': kwargs.get('parameters_spec', {}),
    }

    # ObjectField doesn't allow None nor dict, so if we want to simulate a
    # non-eager object loading, the field should not be referenced at all.
    strategy_data.update(_load_relationships(models.Strategy, kwargs))

    return strategy_data


def get_test_service(**kwargs):
    return {
        'id': kwargs.get('id', 1),
        'name': kwargs.get('name', 'watcher-service'),
        'host': kwargs.get('host', 'controller'),
        'last_seen_up': kwargs.get(
            'last_seen_up',
            timeutils.parse_isotime('2016-09-22T08:32:06').replace(tzinfo=None)
        ),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }


def create_test_service(**kwargs):
    """Create test service entry in DB and return Service DB object.

    Function to be used to create test Service objects in the database.
    :param kwargs: kwargs with overriding values for service's attributes.
    :returns: Test Service DB object.
    """
    service = get_test_service(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.create_service(service)


def create_test_strategy(**kwargs):
    """Create test strategy entry in DB and return Strategy DB object.

    Function to be used to create test Strategy objects in the database.
    :param kwargs: kwargs with overriding values for strategy's attributes.
    :returns: Test Strategy DB object.
    """
    strategy = get_test_strategy(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.create_strategy(strategy)


def get_test_efficacy_indicator(**kwargs):
    return {
        'id': kwargs.get('id', 1),
        'uuid': kwargs.get('uuid', '202cfcf9-811c-411a-8a35-d8351f64eb24'),
        'name': kwargs.get('name', 'test_indicator'),
        'description': kwargs.get('description', 'Test indicator'),
        'unit': kwargs.get('unit', '%'),
        'value': kwargs.get('value', 0),
        'action_plan_id': kwargs.get('action_plan_id', 1),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }


def create_test_efficacy_indicator(**kwargs):
    """Create and return a test efficacy indicator entry in DB.

    Function to be used to create test EfficacyIndicator objects in the DB.
    :param kwargs: kwargs for overriding the values of the attributes
    :returns: Test EfficacyIndicator DB object.
    """
    efficacy_indicator = get_test_efficacy_indicator(**kwargs)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kwargs:
        del efficacy_indicator['id']
    dbapi = db_api.get_instance()
    return dbapi.create_efficacy_indicator(efficacy_indicator)


def get_test_action_desc(**kwargs):
    return {
        'id': kwargs.get('id', 1),
        'action_type': kwargs.get('action_type', 'nop'),
        'description': kwargs.get('description', 'Logging a NOP message'),
        'created_at': kwargs.get('created_at'),
        'updated_at': kwargs.get('updated_at'),
        'deleted_at': kwargs.get('deleted_at'),
    }


def create_test_action_desc(**kwargs):
    """Create test action description entry in DB and return ActionDescription.

    Function to be used to create test ActionDescription objects in the DB.
    :param kwargs: kwargs with overriding values for service's attributes.
    :returns: Test ActionDescription DB object.
    """
    action_desc = get_test_action_desc(**kwargs)
    dbapi = db_api.get_instance()
    return dbapi.create_action_description(action_desc)



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 European Organization for Nuclear Research (CERN)
#
# Authors: Corne Lukken <info@dantalion.nl>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import cfg
from oslo_log import log

from watcher.common import exception
from watcher.decision_engine.datasources.grafana_translator import \
    base as base_translator
from watcher.tests import base

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class TestGrafanaTranslatorBase(base.BaseTestCase):
    """Base class for all GrafanaTranslator test classes

    Objects under test are preceded with t_ and mocked objects are preceded
    with m_ , additionally, patched objects are preceded with p_ no object
    under test should be created in setUp this can influence the results.
    """

    def setUp(self):
        super(TestGrafanaTranslatorBase, self).setUp()

        """Basic valid reference data"""
        self.reference_data = {
            'metric': 'host_cpu_usage',
            'db': 'production',
            'attribute': 'hostname',
            'query': 'SHOW all_base FROM belong_to_us',
            'resource': mock.Mock(hostname='hyperion'),
            'resource_type': 'compute_node',
            'period': '120',
            'aggregate': 'mean',
            'granularity': None
        }


class TestBaseGrafanaTranslator(TestGrafanaTranslatorBase):
    """Test the GrafanaTranslator base class

    Objects under test are preceded with t_ and mocked objects are preceded
    with m_ , additionally, patched objects are preceded with p_ no object
    under test should be created in setUp this can influence the results.
    """

    def setUp(self):
        super(TestBaseGrafanaTranslator, self).setUp()

    def test_validate_data(self):
        """Initialize InfluxDBGrafanaTranslator and check data validation"""

        t_base_translator = base_translator.BaseGrafanaTranslator(
            data=self.reference_data)

        self.assertIsInstance(t_base_translator,
                              base_translator.BaseGrafanaTranslator)

    def test_validate_data_error(self):
        """Initialize InfluxDBGrafanaTranslator and check data validation"""

        self.assertRaises(exception.InvalidParameter,
                          base_translator.BaseGrafanaTranslator,
                          data=[])

    def test_extract_attribute(self):
        """Test that an attribute can be extracted from an object"""

        m_object = mock.Mock(hostname='test')

        t_base_translator = base_translator.BaseGrafanaTranslator(
            data=self.reference_data)

        self.assertEqual('test', t_base_translator._extract_attribute(
            m_object, 'hostname'))

    def test_extract_attribute_error(self):
        """Test error on attempt to extract none existing attribute"""

        m_object = mock.Mock(hostname='test')
        m_object.test = mock.PropertyMock(side_effect=AttributeError)

        t_base_translator = base_translator.BaseGrafanaTranslator(
            data=self.reference_data)

        self.assertRaises(AttributeError, t_base_translator._extract_attribute(
            m_object, 'test'))



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

import iso8601

from watcher.db.sqlalchemy import api as db_api
from watcher import objects
from watcher.tests.db import base
from watcher.tests.db import utils


class TestServiceObject(base.DbTestCase):

    def setUp(self):
        super(TestServiceObject, self).setUp()
        self.fake_service = utils.get_test_service(
            created_at=datetime.datetime.utcnow())

    @mock.patch.object(db_api.Connection, 'get_service_by_id')
    def test_get_by_id(self, mock_get_service):
        service_id = self.fake_service['id']
        mock_get_service.return_value = self.fake_service
        service = objects.Service.get(self.context, service_id)
        mock_get_service.assert_called_once_with(self.context, service_id)
        self.assertEqual(self.context, service._context)

    @mock.patch.object(db_api.Connection, 'get_service_list')
    def test_list(self, mock_get_list):
        mock_get_list.return_value = [self.fake_service]
        services = objects.Service.list(self.context)
        self.assertEqual(1, mock_get_list.call_count, 1)
        self.assertEqual(1, len(services))
        self.assertIsInstance(services[0], objects.Service)
        self.assertEqual(self.context, services[0]._context)

    @mock.patch.object(db_api.Connection, 'create_service')
    def test_create(self, mock_create_service):
        mock_create_service.return_value = self.fake_service
        service = objects.Service(self.context, **self.fake_service)

        service.create()
        expected_service = self.fake_service.copy()
        expected_service['created_at'] = expected_service[
            'created_at'].replace(tzinfo=iso8601.UTC)

        mock_create_service.assert_called_once_with(expected_service)
        self.assertEqual(self.context, service._context)

    @mock.patch.object(db_api.Connection, 'update_service')
    @mock.patch.object(db_api.Connection, 'get_service_by_id')
    def test_save(self, mock_get_service, mock_update_service):
        mock_get_service.return_value = self.fake_service
        fake_saved_service = self.fake_service.copy()
        fake_saved_service['updated_at'] = datetime.datetime.utcnow()
        mock_update_service.return_value = fake_saved_service
        _id = self.fake_service['id']
        service = objects.Service.get(self.context, _id)
        service.name = 'UPDATED NAME'
        service.save()

        mock_get_service.assert_called_once_with(self.context, _id)
        mock_update_service.assert_called_once_with(
            _id, {'name': 'UPDATED NAME'})
        self.assertEqual(self.context, service._context)

    @mock.patch.object(db_api.Connection, 'get_service_by_id')
    def test_refresh(self, mock_get_service):
        returns = [dict(self.fake_service, name="first name"),
                   dict(self.fake_service, name="second name")]
        mock_get_service.side_effect = returns
        _id = self.fake_service['id']
        expected = [mock.call(self.context, _id),
                    mock.call(self.context, _id)]
        service = objects.Service.get(self.context, _id)
        self.assertEqual("first name", service.name)
        service.refresh()
        self.assertEqual("second name", service.name)
        self.assertEqual(expected, mock_get_service.call_args_list)
        self.assertEqual(self.context, service._context)

    @mock.patch.object(db_api.Connection, 'soft_delete_service')
    @mock.patch.object(db_api.Connection, 'get_service_by_id')
    def test_soft_delete(self, mock_get_service, mock_soft_delete):
        mock_get_service.return_value = self.fake_service
        fake_deleted_service = self.fake_service.copy()
        fake_deleted_service['deleted_at'] = datetime.datetime.utcnow()
        mock_soft_delete.return_value = fake_deleted_service

        expected_service = fake_deleted_service.copy()
        expected_service['created_at'] = expected_service[
            'created_at'].replace(tzinfo=iso8601.UTC)
        expected_service['deleted_at'] = expected_service[
            'deleted_at'].replace(tzinfo=iso8601.UTC)

        _id = self.fake_service['id']
        service = objects.Service.get(self.context, _id)
        service.soft_delete()
        mock_get_service.assert_called_once_with(self.context, _id)
        mock_soft_delete.assert_called_once_with(_id)
        self.assertEqual(self.context, service._context)
        self.assertEqual(expected_service, service.as_dict())



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <Vincent.FRANCOISE@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class NotificationEndpoint(object, metaclass=abc.ABCMeta):

    def __init__(self, collector):
        super(NotificationEndpoint, self).__init__()
        self.collector = collector
        self._notifier = None

    @abc.abstractproperty
    def filter_rule(self):
        """Notification Filter"""
        raise NotImplementedError()

    @property
    def cluster_data_model(self):
        return self.collector.cluster_data_model



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from http import HTTPStatus

from watcher.api.controllers.v1 import versions
from watcher.tests.api import base as api_base


SERVICE_TYPE = 'infra-optim'
H_MIN_VER = 'openstack-api-minimum-version'
H_MAX_VER = 'openstack-api-maximum-version'
H_RESP_VER = 'openstack-api-version'
MIN_VER = versions.min_version_string()
MAX_VER = versions.max_version_string()


class TestMicroversions(api_base.FunctionalTest):

    controller_list_response = [
        'scoring_engines', 'audit_templates', 'audits', 'actions',
        'action_plans', 'services']

    def setUp(self):
        super(TestMicroversions, self).setUp()

    def test_wrong_major_version(self):
        response = self.get_json(
            '/',
            headers={'OpenStack-API-Version': ' '.join([SERVICE_TYPE,
                                                        '10'])},
            expect_errors=True, return_json=False)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(HTTPStatus.NOT_ACCEPTABLE, response.status_int)
        expected_error_msg = ('Invalid value for'
                              ' OpenStack-API-Version header')
        self.assertTrue(response.json['error_message'])
        self.assertIn(expected_error_msg, response.json['error_message'])

    def test_extend_initial_version_with_micro(self):
        response = self.get_json(
            '/',
            headers={'OpenStack-API-Version': ' '.join([SERVICE_TYPE,
                                                        '1'])},
            return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER],
                         ' '.join([SERVICE_TYPE, MIN_VER]))
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_without_microversion(self):
        response = self.get_json('/', return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER],
                         ' '.join([SERVICE_TYPE, MIN_VER]))
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_new_client_new_api(self):
        response = self.get_json(
            '/',
            headers={'OpenStack-API-Version': ' '.join([SERVICE_TYPE,
                                                        '1.1'])},
            return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER],
                         ' '.join([SERVICE_TYPE, '1.1']))
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_latest_microversion(self):
        response = self.get_json(
            '/',
            headers={'OpenStack-API-Version': ' '.join([SERVICE_TYPE,
                                                        'latest'])},
            return_json=False)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        self.assertEqual(response.headers[H_RESP_VER],
                         ' '.join([SERVICE_TYPE, MAX_VER]))
        self.assertTrue(all(x in response.json.keys() for x in
                            self.controller_list_response))

    def test_unsupported_version(self):
        response = self.get_json(
            '/',
            headers={'OpenStack-API-Version': ' '.join([SERVICE_TYPE,
                                                        '1.999'])},
            expect_errors=True)
        self.assertEqual(HTTPStatus.NOT_ACCEPTABLE, response.status_int)
        self.assertEqual(response.headers[H_MIN_VER], MIN_VER)
        self.assertEqual(response.headers[H_MAX_VER], MAX_VER)
        expected_error_msg = ('Version 1.999 was requested but the minor '
                              'version is not supported by this service. '
                              'The supported version range is')
        self.assertTrue(response.json['error_message'])
        self.assertIn(expected_error_msg, response.json['error_message'])



# -*- encoding: utf-8 -*-
# Copyright (c) 2018 Servionica
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.decision_engine.strategy.strategies import base as strategy_base
from watcher.tests import base


class TestStrategyEndpoint(base.BaseTestCase):

    def test_collect_metrics(self):
        datasource = mock.MagicMock()
        datasource.list_metrics.return_value = ["m1", "m2"]
        datasource.METRIC_MAP = {"metric1": "m1", "metric2": "m2",
                                 "metric3": "m3"}
        strategy = mock.MagicMock()
        strategy.DATASOURCE_METRICS = ["metric1", "metric2", "metric3"]
        strategy.config.datasource = "gnocchi"
        se = strategy_base.StrategyEndpoint(mock.MagicMock())
        result = se._collect_metrics(strategy, datasource)
        expected_result = {'type': 'Metrics',
                           'state': [{"m1": "available"},
                                     {"m2": "available"},
                                     {"m3": "not available"}],
                           'mandatory': False, 'comment': ''}
        self.assertEqual(expected_result, result)

    def test_get_datasource_status(self):
        strategy = mock.MagicMock()
        datasource = mock.MagicMock()
        datasource.NAME = 'gnocchi'
        datasource.check_availability.return_value = "available"
        se = strategy_base.StrategyEndpoint(mock.MagicMock())
        result = se._get_datasource_status(strategy, datasource)
        expected_result = {'type': 'Datasource',
                           'state': "gnocchi: available",
                           'mandatory': True, 'comment': ''}
        self.assertEqual(expected_result, result)

    def test_get_cdm(self):
        strategy = mock.MagicMock()
        strategy.compute_model = mock.MagicMock()
        del strategy.storage_model
        strategy.baremetal_model = mock.MagicMock()
        se = strategy_base.StrategyEndpoint(mock.MagicMock())
        result = se._get_cdm(strategy)
        expected_result = {'type': 'CDM',
                           'state': [{"compute_model": "available"},
                                     {"storage_model": "not available"},
                                     {"baremetal_model": "available"}],
                           'mandatory': True, 'comment': ''}
        self.assertEqual(expected_result, result)



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel Innovation and Research Ireland Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Openstack implementation of the cluster graph.
"""

import ast
from lxml import etree
import networkx as nx
from oslo_concurrency import lockutils
from oslo_log import log

from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.model import base
from watcher.decision_engine.model import element

LOG = log.getLogger(__name__)


class ModelRoot(nx.DiGraph, base.Model):
    """Cluster graph for an Openstack cluster."""

    def __init__(self, stale=False):
        super(ModelRoot, self).__init__()
        self.stale = stale

    def __nonzero__(self):
        return not self.stale

    __bool__ = __nonzero__

    @staticmethod
    def assert_node(obj):
        if not isinstance(obj, element.ComputeNode):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid: %s") % type(obj))

    @staticmethod
    def assert_instance(obj):
        if not isinstance(obj, element.Instance):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid"))

    @lockutils.synchronized("model_root")
    def add_node(self, node):
        self.assert_node(node)
        super(ModelRoot, self).add_node(node.uuid, attr=node)

    @lockutils.synchronized("model_root")
    def remove_node(self, node):
        self.assert_node(node)
        try:
            super(ModelRoot, self).remove_node(node.uuid)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.ComputeNodeNotFound(name=node.uuid)

    @lockutils.synchronized("model_root")
    def add_instance(self, instance):
        self.assert_instance(instance)
        try:
            super(ModelRoot, self).add_node(instance.uuid, attr=instance)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.InstanceNotFound(name=instance.uuid)

    @lockutils.synchronized("model_root")
    def remove_instance(self, instance):
        self.assert_instance(instance)
        super(ModelRoot, self).remove_node(instance.uuid)

    @lockutils.synchronized("model_root")
    def map_instance(self, instance, node):
        """Map a newly created instance to a node

        :param instance: :py:class:`~.instance.Instance` object or instance
           UUID
        :type instance: str or :py:class:`~.instance.Instance`
        :param node: :py:class:`~.node.ComputeNode` object or node UUID
        :type node: str or :py:class:`~.instance.Instance`
        """
        if isinstance(instance, str):
            instance = self.get_instance_by_uuid(instance)
        if isinstance(node, str):
            node = self.get_node_by_uuid(node)
        self.assert_node(node)
        self.assert_instance(instance)

        self.add_edge(instance.uuid, node.uuid)

    @lockutils.synchronized("model_root")
    def unmap_instance(self, instance, node):
        if isinstance(instance, str):
            instance = self.get_instance_by_uuid(instance)
        if isinstance(node, str):
            node = self.get_node_by_uuid(node)

        self.remove_edge(instance.uuid, node.uuid)

    def delete_instance(self, instance, node=None):
        self.assert_instance(instance)
        self.remove_instance(instance)

    @lockutils.synchronized("model_root")
    def migrate_instance(self, instance, source_node, destination_node):
        """Migrate single instance from source_node to destination_node

        :param instance:
        :param source_node:
        :param destination_node:
        :return:
        """
        self.assert_instance(instance)
        self.assert_node(source_node)
        self.assert_node(destination_node)

        if source_node == destination_node:
            return False

        # unmap
        self.remove_edge(instance.uuid, source_node.uuid)
        # map
        self.add_edge(instance.uuid, destination_node.uuid)
        return True

    @lockutils.synchronized("model_root")
    def get_all_compute_nodes(self):
        return {uuid: cn['attr'] for uuid, cn in self.nodes(data=True)
                if isinstance(cn['attr'], element.ComputeNode)}

    @lockutils.synchronized("model_root")
    def get_node_by_uuid(self, uuid):
        try:
            return self._get_by_uuid(uuid)
        except exception.ComputeResourceNotFound:
            raise exception.ComputeNodeNotFound(name=uuid)

    @lockutils.synchronized("model_root")
    def get_node_by_name(self, name):
        try:
            node_list = [cn['attr'] for uuid, cn in self.nodes(data=True)
                         if (isinstance(cn['attr'], element.ComputeNode) and
                         cn['attr']['hostname'] == name)]
            if node_list:
                return node_list[0]
            else:
                raise exception.ComputeResourceNotFound
        except exception.ComputeResourceNotFound:
            raise exception.ComputeNodeNotFound(name=name)

    @lockutils.synchronized("model_root")
    def get_instance_by_uuid(self, uuid):
        try:
            return self._get_by_uuid(uuid)
        except exception.ComputeResourceNotFound:
            raise exception.InstanceNotFound(name=uuid)

    def _get_by_uuid(self, uuid):
        try:
            return self.nodes[uuid]['attr']
        except Exception as exc:
            LOG.exception(exc)
            raise exception.ComputeResourceNotFound(name=uuid)

    @lockutils.synchronized("model_root")
    def get_node_by_instance_uuid(self, instance_uuid):
        instance = self._get_by_uuid(instance_uuid)
        for node_uuid in self.neighbors(instance.uuid):
            node = self._get_by_uuid(node_uuid)
            if isinstance(node, element.ComputeNode):
                return node
        raise exception.InstanceNotMapped(uuid=instance_uuid)

    @lockutils.synchronized("model_root")
    def get_all_instances(self):
        return {uuid: inst['attr'] for uuid, inst in self.nodes(data=True)
                if isinstance(inst['attr'], element.Instance)}

    @lockutils.synchronized("model_root")
    def get_node_instances(self, node):
        self.assert_node(node)
        node_instances = []
        for instance_uuid in self.predecessors(node.uuid):
            instance = self._get_by_uuid(instance_uuid)
            if isinstance(instance, element.Instance):
                node_instances.append(instance)

        return node_instances

    def get_node_used_resources(self, node):
        vcpu_used = 0
        memory_used = 0
        disk_used = 0
        for instance in self.get_node_instances(node):
            vcpu_used += instance.vcpus
            memory_used += instance.memory
            disk_used += instance.disk

        return dict(vcpu=vcpu_used, memory=memory_used, disk=disk_used)

    def get_node_free_resources(self, node):
        resources_used = self.get_node_used_resources(node)
        vcpu_free = node.vcpu_capacity-resources_used.get('vcpu')
        memory_free = node.memory_mb_capacity-resources_used.get('memory')
        disk_free = node.disk_gb_capacity-resources_used.get('disk')

        return dict(vcpu=vcpu_free, memory=memory_free, disk=disk_free)

    def to_string(self):
        return self.to_xml()

    def to_xml(self):
        root = etree.Element("ModelRoot")
        # Build compute node tree
        for cn in sorted(self.get_all_compute_nodes().values(),
                         key=lambda cn: cn.uuid):
            compute_node_el = cn.as_xml_element()

            # Build mapped instance tree
            node_instances = self.get_node_instances(cn)
            for instance in sorted(node_instances, key=lambda x: x.uuid):
                instance_el = instance.as_xml_element()
                compute_node_el.append(instance_el)

            root.append(compute_node_el)

        # Build unmapped instance tree (i.e. not assigned to any compute node)
        for instance in sorted(self.get_all_instances().values(),
                               key=lambda inst: inst.uuid):
            try:
                self.get_node_by_instance_uuid(instance.uuid)
            except exception.ComputeResourceNotFound:
                root.append(instance.as_xml_element())

        return etree.tostring(root, pretty_print=True).decode('utf-8')

    def to_list(self):
        ret_list = []
        for cn in sorted(self.get_all_compute_nodes().values(),
                         key=lambda cn: cn.uuid):
            in_dict = {}
            for field in cn.fields:
                new_name = "node_"+str(field)
                in_dict[new_name] = cn[field]
            node_instances = self.get_node_instances(cn)
            if not node_instances:
                deep_in_dict = in_dict.copy()
                ret_list.append(deep_in_dict)
                continue
            for instance in sorted(node_instances, key=lambda x: x.uuid):
                for field in instance.fields:
                    new_name = "server_"+str(field)
                    in_dict[new_name] = instance[field]
                if in_dict != {}:
                    deep_in_dict = in_dict.copy()
                    ret_list.append(deep_in_dict)
        return ret_list

    @classmethod
    def from_xml(cls, data):
        model = cls()

        root = etree.fromstring(data)
        for cn in root.findall('.//ComputeNode'):
            node = element.ComputeNode(**cn.attrib)
            model.add_node(node)

        for inst in root.findall('.//Instance'):
            instance = element.Instance(**inst.attrib)
            instance.watcher_exclude = ast.literal_eval(
                inst.attrib["watcher_exclude"])
            model.add_instance(instance)

            parent = inst.getparent()
            if parent.tag == 'ComputeNode':
                node = model.get_node_by_uuid(parent.get('uuid'))
                model.map_instance(instance, node)
            else:
                model.add_instance(instance)

        return model

    @classmethod
    def is_isomorphic(cls, G1, G2):
        def node_match(node1, node2):
            return node1['attr'].as_dict() == node2['attr'].as_dict()
        return nx.algorithms.isomorphism.isomorph.is_isomorphic(
            G1, G2, node_match=node_match)


class StorageModelRoot(nx.DiGraph, base.Model):
    """Cluster graph for an Openstack cluster."""

    def __init__(self, stale=False):
        super(StorageModelRoot, self).__init__()
        self.stale = stale

    def __nonzero__(self):
        return not self.stale

    __bool__ = __nonzero__

    @staticmethod
    def assert_node(obj):
        if not isinstance(obj, element.StorageNode):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid: %s") % type(obj))

    @staticmethod
    def assert_pool(obj):
        if not isinstance(obj, element.Pool):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid: %s") % type(obj))

    @staticmethod
    def assert_volume(obj):
        if not isinstance(obj, element.Volume):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid: %s") % type(obj))

    @lockutils.synchronized("storage_model")
    def add_node(self, node):
        self.assert_node(node)
        super(StorageModelRoot, self).add_node(node.host, attr=node)

    @lockutils.synchronized("storage_model")
    def add_pool(self, pool):
        self.assert_pool(pool)
        super(StorageModelRoot, self).add_node(pool.name, attr=pool)

    @lockutils.synchronized("storage_model")
    def remove_node(self, node):
        self.assert_node(node)
        try:
            super(StorageModelRoot, self).remove_node(node.host)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.StorageNodeNotFound(name=node.host)

    @lockutils.synchronized("storage_model")
    def remove_pool(self, pool):
        self.assert_pool(pool)
        try:
            super(StorageModelRoot, self).remove_node(pool.name)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.PoolNotFound(name=pool.name)

    @lockutils.synchronized("storage_model")
    def map_pool(self, pool, node):
        """Map a newly created pool to a node

        :param pool: :py:class:`~.node.Pool` object or pool name
        :param node: :py:class:`~.node.StorageNode` object or node host
        """
        if isinstance(pool, str):
            pool = self.get_pool_by_pool_name(pool)
        if isinstance(node, str):
            node = self.get_node_by_name(node)
        self.assert_node(node)
        self.assert_pool(pool)

        self.add_edge(pool.name, node.host)

    @lockutils.synchronized("storage_model")
    def unmap_pool(self, pool, node):
        """Unmap a pool from a node

        :param pool: :py:class:`~.node.Pool` object or pool name
        :param node: :py:class:`~.node.StorageNode` object or node name
        """
        if isinstance(pool, str):
            pool = self.get_pool_by_pool_name(pool)
        if isinstance(node, str):
            node = self.get_node_by_name(node)

        self.remove_edge(pool.name, node.host)

    @lockutils.synchronized("storage_model")
    def add_volume(self, volume):
        self.assert_volume(volume)
        super(StorageModelRoot, self).add_node(volume.uuid, attr=volume)

    @lockutils.synchronized("storage_model")
    def remove_volume(self, volume):
        self.assert_volume(volume)
        try:
            super(StorageModelRoot, self).remove_node(volume.uuid)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.VolumeNotFound(name=volume.uuid)

    @lockutils.synchronized("storage_model")
    def map_volume(self, volume, pool):
        """Map a newly created volume to a pool

        :param volume: :py:class:`~.volume.Volume` object or volume UUID
        :param pool: :py:class:`~.node.Pool` object or pool name
        """
        if isinstance(volume, str):
            volume = self.get_volume_by_uuid(volume)
        if isinstance(pool, str):
            pool = self.get_pool_by_pool_name(pool)
        self.assert_pool(pool)
        self.assert_volume(volume)

        self.add_edge(volume.uuid, pool.name)

    @lockutils.synchronized("storage_model")
    def unmap_volume(self, volume, pool):
        """Unmap a volume from a pool

        :param volume: :py:class:`~.volume.Volume` object or volume UUID
        :param pool: :py:class:`~.node.Pool` object or pool name
        """
        if isinstance(volume, str):
            volume = self.get_volume_by_uuid(volume)
        if isinstance(pool, str):
            pool = self.get_pool_by_pool_name(pool)

        self.remove_edge(volume.uuid, pool.name)

    def delete_volume(self, volume):
        self.assert_volume(volume)
        self.remove_volume(volume)

    @lockutils.synchronized("storage_model")
    def get_all_storage_nodes(self):
        return {host: cn['attr'] for host, cn in self.nodes(data=True)
                if isinstance(cn['attr'], element.StorageNode)}

    @lockutils.synchronized("storage_model")
    def get_node_by_name(self, name):
        try:
            return self._get_by_name(name.split("#")[0])
        except exception.StorageResourceNotFound:
            raise exception.StorageNodeNotFound(name=name)

    @lockutils.synchronized("storage_model")
    def get_pool_by_pool_name(self, name):
        try:
            return self._get_by_name(name)
        except exception.StorageResourceNotFound:
            raise exception.PoolNotFound(name=name)

    @lockutils.synchronized("storage_model")
    def get_volume_by_uuid(self, uuid):
        try:
            return self._get_by_uuid(uuid)
        except exception.StorageResourceNotFound:
            raise exception.VolumeNotFound(name=uuid)

    def _get_by_uuid(self, uuid):
        try:
            return self.nodes[uuid]['attr']
        except Exception as exc:
            LOG.exception(exc)
            raise exception.StorageResourceNotFound(name=uuid)

    def _get_by_name(self, name):
        try:
            return self.nodes[name]['attr']
        except Exception as exc:
            LOG.exception(exc)
            raise exception.StorageResourceNotFound(name=name)

    @lockutils.synchronized("storage_model")
    def get_node_by_pool_name(self, pool_name):
        pool = self._get_by_name(pool_name)
        for node_name in self.neighbors(pool.name):
            node = self._get_by_name(node_name)
            if isinstance(node, element.StorageNode):
                return node
        raise exception.StorageNodeNotFound(name=pool_name)

    @lockutils.synchronized("storage_model")
    def get_node_pools(self, node):
        self.assert_node(node)
        node_pools = []
        for pool_name in self.predecessors(node.host):
            pool = self._get_by_name(pool_name)
            if isinstance(pool, element.Pool):
                node_pools.append(pool)

        return node_pools

    @lockutils.synchronized("storage_model")
    def get_pool_by_volume(self, volume):
        self.assert_volume(volume)
        volume = self._get_by_uuid(volume.uuid)
        for p in self.neighbors(volume.uuid):
            pool = self._get_by_name(p)
            if isinstance(pool, element.Pool):
                return pool
        raise exception.PoolNotFound(name=volume.uuid)

    @lockutils.synchronized("storage_model")
    def get_all_volumes(self):
        return {name: vol['attr'] for name, vol in self.nodes(data=True)
                if isinstance(vol['attr'], element.Volume)}

    @lockutils.synchronized("storage_model")
    def get_pool_volumes(self, pool):
        self.assert_pool(pool)
        volumes = []
        for vol in self.predecessors(pool.name):
            volume = self._get_by_uuid(vol)
            if isinstance(volume, element.Volume):
                volumes.append(volume)

        return volumes

    def to_string(self):
        return self.to_xml()

    def to_xml(self):
        root = etree.Element("ModelRoot")
        # Build storage node tree
        for cn in sorted(self.get_all_storage_nodes().values(),
                         key=lambda cn: cn.host):
            storage_node_el = cn.as_xml_element()
            # Build mapped pool tree
            node_pools = self.get_node_pools(cn)
            for pool in sorted(node_pools, key=lambda x: x.name):
                pool_el = pool.as_xml_element()
                storage_node_el.append(pool_el)
                # Build mapped volume tree
                pool_volumes = self.get_pool_volumes(pool)
                for volume in sorted(pool_volumes, key=lambda x: x.uuid):
                    volume_el = volume.as_xml_element()
                    pool_el.append(volume_el)

            root.append(storage_node_el)

        # Build unmapped volume tree (i.e. not assigned to any pool)
        for volume in sorted(self.get_all_volumes().values(),
                             key=lambda vol: vol.uuid):
            try:
                self.get_pool_by_volume(volume)
            except (exception.VolumeNotFound, exception.PoolNotFound):
                root.append(volume.as_xml_element())

        return etree.tostring(root, pretty_print=True).decode('utf-8')

    @classmethod
    def from_xml(cls, data):
        model = cls()

        root = etree.fromstring(data)
        for cn in root.findall('.//StorageNode'):
            ndata = {}
            for attr, val in cn.items():
                ndata[attr] = val
            volume_type = ndata.get('volume_type')
            if volume_type:
                ndata['volume_type'] = [volume_type]
            node = element.StorageNode(**ndata)
            model.add_node(node)

        for p in root.findall('.//Pool'):
            pool = element.Pool(**p.attrib)
            model.add_pool(pool)

            parent = p.getparent()
            if parent.tag == 'StorageNode':
                node = model.get_node_by_name(parent.get('host'))
                model.map_pool(pool, node)
            else:
                model.add_pool(pool)

        for vol in root.findall('.//Volume'):
            volume = element.Volume(**vol.attrib)
            model.add_volume(volume)

            parent = vol.getparent()
            if parent.tag == 'Pool':
                pool = model.get_pool_by_pool_name(parent.get('name'))
                model.map_volume(volume, pool)
            else:
                model.add_volume(volume)

        return model

    @classmethod
    def is_isomorphic(cls, G1, G2):
        return nx.algorithms.isomorphism.isomorph.is_isomorphic(
            G1, G2)


class BaremetalModelRoot(nx.DiGraph, base.Model):

    """Cluster graph for an Openstack cluster: Baremetal Cluster."""

    def __init__(self, stale=False):
        super(BaremetalModelRoot, self).__init__()
        self.stale = stale

    def __nonzero__(self):
        return not self.stale

    __bool__ = __nonzero__

    @staticmethod
    def assert_node(obj):
        if not isinstance(obj, element.IronicNode):
            raise exception.IllegalArgumentException(
                message=_("'obj' argument type is not valid: %s") % type(obj))

    @lockutils.synchronized("baremetal_model")
    def add_node(self, node):
        self.assert_node(node)
        super(BaremetalModelRoot, self).add_node(node.uuid, attr=node)

    @lockutils.synchronized("baremetal_model")
    def remove_node(self, node):
        self.assert_node(node)
        try:
            super(BaremetalModelRoot, self).remove_node(node.uuid)
        except nx.NetworkXError as exc:
            LOG.exception(exc)
            raise exception.IronicNodeNotFound(uuid=node.uuid)

    @lockutils.synchronized("baremetal_model")
    def get_all_ironic_nodes(self):
        return {uuid: cn['attr'] for uuid, cn in self.nodes(data=True)
                if isinstance(cn['attr'], element.IronicNode)}

    @lockutils.synchronized("baremetal_model")
    def get_node_by_uuid(self, uuid):
        try:
            return self._get_by_uuid(uuid)
        except exception.BaremetalResourceNotFound:
            raise exception.IronicNodeNotFound(uuid=uuid)

    def _get_by_uuid(self, uuid):
        try:
            return self.nodes[uuid]['attr']
        except Exception as exc:
            LOG.exception(exc)
        raise exception.BaremetalResourceNotFound(name=uuid)

    def to_string(self):
        return self.to_xml()

    def to_xml(self):
        root = etree.Element("ModelRoot")
        # Build Ironic node tree
        for cn in sorted(self.get_all_ironic_nodes().values(),
                         key=lambda cn: cn.uuid):
            ironic_node_el = cn.as_xml_element()
            root.append(ironic_node_el)

        return etree.tostring(root, pretty_print=True).decode('utf-8')

    @classmethod
    def from_xml(cls, data):
        model = cls()

        root = etree.fromstring(data)
        for cn in root.findall('.//IronicNode'):
            node = element.IronicNode(**cn.attrib)
            model.add_node(node)

        return model

    @classmethod
    def is_isomorphic(cls, G1, G2):
        return nx.algorithms.isomorphism.isomorph.is_isomorphic(
            G1, G2)



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import collections

from oslo_log import log

from watcher.common import context
from watcher.decision_engine.loading import default
from watcher.decision_engine.scoring import scoring_factory
from watcher import objects

LOG = log.getLogger(__name__)

GoalMapping = collections.namedtuple(
    'GoalMapping', ['name', 'display_name', 'efficacy_specification'])
StrategyMapping = collections.namedtuple(
    'StrategyMapping',
    ['name', 'goal_name', 'display_name', 'parameters_spec'])
ScoringEngineMapping = collections.namedtuple(
    'ScoringEngineMapping',
    ['name', 'description', 'metainfo'])

IndicatorSpec = collections.namedtuple(
    'IndicatorSpec', ['name', 'description', 'unit', 'schema'])


class Syncer(object):
    """Syncs all available goals and strategies with the Watcher DB"""

    def __init__(self):
        self.ctx = context.make_context()
        self.discovered_map = None

        self._available_goals = None
        self._available_goals_map = None

        self._available_strategies = None
        self._available_strategies_map = None

        self._available_scoringengines = None
        self._available_scoringengines_map = None

        # This goal mapping maps stale goal IDs to the synced goal
        self.goal_mapping = dict()
        # This strategy mapping maps stale strategy IDs to the synced goal
        self.strategy_mapping = dict()
        # Maps stale scoring engine IDs to the synced scoring engines
        self.se_mapping = dict()

        self.stale_audit_templates_map = {}
        self.stale_audits_map = {}
        self.stale_action_plans_map = {}

    @property
    def available_goals(self):
        """Goals loaded from DB"""
        if self._available_goals is None:
            self._available_goals = objects.Goal.list(self.ctx)
        return self._available_goals

    @property
    def available_strategies(self):
        """Strategies loaded from DB"""
        if self._available_strategies is None:
            self._available_strategies = objects.Strategy.list(self.ctx)
            goal_ids = [g.id for g in self.available_goals]
            stale_strategies = [s for s in self._available_strategies
                                if s.goal_id not in goal_ids]
            for s in stale_strategies:
                LOG.info("Can't find Goal id %d of strategy %s",
                         s.goal_id, s.name)
                s.soft_delete()
                self._available_strategies.remove(s)
        return self._available_strategies

    @property
    def available_scoringengines(self):
        """Scoring Engines loaded from DB"""
        if self._available_scoringengines is None:
            self._available_scoringengines = (objects.ScoringEngine
                                              .list(self.ctx))
        return self._available_scoringengines

    @property
    def available_goals_map(self):
        """Mapping of goals loaded from DB"""
        if self._available_goals_map is None:
            self._available_goals_map = {
                GoalMapping(
                    name=g.name,
                    display_name=g.display_name,
                    efficacy_specification=tuple(
                        IndicatorSpec(**item)
                        for item in g.efficacy_specification)): g
                for g in self.available_goals
            }
        return self._available_goals_map

    @property
    def available_strategies_map(self):
        if self._available_strategies_map is None:
            goals_map = {g.id: g.name for g in self.available_goals}
            self._available_strategies_map = {
                StrategyMapping(
                    name=s.name, goal_name=goals_map[s.goal_id],
                    display_name=s.display_name,
                    parameters_spec=str(s.parameters_spec)): s
                for s in self.available_strategies
            }
        return self._available_strategies_map

    @property
    def available_scoringengines_map(self):
        if self._available_scoringengines_map is None:
            self._available_scoringengines_map = {
                ScoringEngineMapping(
                    name=s.id, description=s.description,
                    metainfo=s.metainfo): s
                for s in self.available_scoringengines
            }
        return self._available_scoringengines_map

    def sync(self):
        self.discovered_map = self._discover()
        goals_map = self.discovered_map["goals"]
        strategies_map = self.discovered_map["strategies"]
        scoringengines_map = self.discovered_map["scoringengines"]

        for goal_name, goal_map in goals_map.items():
            if goal_map in self.available_goals_map:
                LOG.info("Goal %s already exists", goal_name)
                continue

            self.goal_mapping.update(self._sync_goal(goal_map))

        for strategy_name, strategy_map in strategies_map.items():
            if (strategy_map in self.available_strategies_map and
                    strategy_map.goal_name not in
                    [g.name for g in self.goal_mapping.values()]):
                LOG.info("Strategy %s already exists", strategy_name)
                continue

            self.strategy_mapping.update(self._sync_strategy(strategy_map))

        for se_name, se_map in scoringengines_map.items():
            if se_map in self.available_scoringengines_map:
                LOG.info("Scoring Engine %s already exists",
                         se_name)
                continue

            self.se_mapping.update(self._sync_scoringengine(se_map))

        self._sync_objects()
        self._soft_delete_removed_scoringengines()

    def _sync_goal(self, goal_map):
        goal_name = goal_map.name
        goal_mapping = dict()
        # Goals that are matching by name with the given discovered goal name
        matching_goals = [g for g in self.available_goals
                          if g.name == goal_name]
        stale_goals = self._soft_delete_stale_goals(goal_map, matching_goals)

        if stale_goals or not matching_goals:
            goal = objects.Goal(self.ctx)
            goal.name = goal_name
            goal.display_name = goal_map.display_name
            goal.efficacy_specification = [
                indicator._asdict()
                for indicator in goal_map.efficacy_specification]
            goal.create()
            LOG.info("Goal %s created", goal_name)

            # Updating the internal states
            self.available_goals_map[goal] = goal_map
            # Map the old goal IDs to the new (equivalent) goal
            for matching_goal in matching_goals:
                goal_mapping[matching_goal.id] = goal

        return goal_mapping

    def _sync_strategy(self, strategy_map):
        strategy_name = strategy_map.name
        strategy_display_name = strategy_map.display_name
        goal_name = strategy_map.goal_name
        parameters_spec = strategy_map.parameters_spec
        strategy_mapping = dict()

        # Strategies that are matching by name with the given
        # discovered strategy name
        matching_strategies = [s for s in self.available_strategies
                               if s.name == strategy_name]
        stale_strategies = self._soft_delete_stale_strategies(
            strategy_map, matching_strategies)

        if stale_strategies or not matching_strategies:
            strategy = objects.Strategy(self.ctx)
            strategy.name = strategy_name
            strategy.display_name = strategy_display_name
            strategy.goal_id = objects.Goal.get_by_name(self.ctx, goal_name).id
            strategy.parameters_spec = parameters_spec
            strategy.create()
            LOG.info("Strategy %s created", strategy_name)

            # Updating the internal states
            self.available_strategies_map[strategy] = strategy_map
            # Map the old strategy IDs to the new (equivalent) strategy
            for matching_strategy in matching_strategies:
                strategy_mapping[matching_strategy.id] = strategy

        return strategy_mapping

    def _sync_scoringengine(self, scoringengine_map):
        scoringengine_name = scoringengine_map.name
        se_mapping = dict()
        # Scoring Engines matching by id with discovered Scoring engine
        matching_scoringengines = [se for se in self.available_scoringengines
                                   if se.name == scoringengine_name]
        stale_scoringengines = self._soft_delete_stale_scoringengines(
            scoringengine_map, matching_scoringengines)

        if stale_scoringengines or not matching_scoringengines:
            scoringengine = objects.ScoringEngine(self.ctx)
            scoringengine.name = scoringengine_name
            scoringengine.description = scoringengine_map.description
            scoringengine.metainfo = scoringengine_map.metainfo
            scoringengine.create()
            LOG.info("Scoring Engine %s created", scoringengine_name)

            # Updating the internal states
            self.available_scoringengines_map[scoringengine] = \
                scoringengine_map
            # Map the old scoring engine names to the new (equivalent) SE
            for matching_scoringengine in matching_scoringengines:
                se_mapping[matching_scoringengine.name] = scoringengine

        return se_mapping

    def _sync_objects(self):
        # First we find audit templates, audits and action plans that are stale
        # because their associated goal or strategy has been modified and we
        # update them in-memory
        self._find_stale_audit_templates_due_to_goal()
        self._find_stale_audit_templates_due_to_strategy()

        self._find_stale_audits_due_to_goal()
        self._find_stale_audits_due_to_strategy()

        self._find_stale_action_plans_due_to_strategy()
        self._find_stale_action_plans_due_to_audit()

        # Then we handle the case where an audit template, an audit or an
        # action plan becomes stale because its related goal does not
        # exist anymore.
        self._soft_delete_removed_goals()
        # Then we handle the case where an audit template, an audit or an
        # action plan becomes stale because its related strategy does not
        # exist anymore.
        self._soft_delete_removed_strategies()

        # Finally, we save into the DB the updated stale audit templates
        # and soft delete stale audits and action plans
        for stale_audit_template in self.stale_audit_templates_map.values():
            stale_audit_template.save()
            LOG.info("Audit Template '%s' synced",
                     stale_audit_template.name)

        for stale_audit in self.stale_audits_map.values():
            stale_audit.save()
            LOG.info("Stale audit '%s' synced and cancelled",
                     stale_audit.uuid)

        for stale_action_plan in self.stale_action_plans_map.values():
            stale_action_plan.save()
            LOG.info("Stale action plan '%s' synced and cancelled",
                     stale_action_plan.uuid)

    def _find_stale_audit_templates_due_to_goal(self):
        for goal_id, synced_goal in self.goal_mapping.items():
            filters = {"goal_id": goal_id}
            stale_audit_templates = objects.AuditTemplate.list(
                self.ctx, filters=filters)

            # Update the goal ID for the stale audit templates (w/o saving)
            for audit_template in stale_audit_templates:
                if audit_template.id not in self.stale_audit_templates_map:
                    audit_template.goal_id = synced_goal.id
                    self.stale_audit_templates_map[audit_template.id] = (
                        audit_template)
                else:
                    self.stale_audit_templates_map[
                        audit_template.id].goal_id = synced_goal.id

    def _find_stale_audit_templates_due_to_strategy(self):
        for strategy_id, synced_strategy in self.strategy_mapping.items():
            filters = {"strategy_id": strategy_id}
            stale_audit_templates = objects.AuditTemplate.list(
                self.ctx, filters=filters)

            # Update strategy IDs for all stale audit templates (w/o saving)
            for audit_template in stale_audit_templates:
                if audit_template.id not in self.stale_audit_templates_map:
                    audit_template.strategy_id = synced_strategy.id
                    self.stale_audit_templates_map[audit_template.id] = (
                        audit_template)
                else:
                    self.stale_audit_templates_map[
                        audit_template.id].strategy_id = synced_strategy.id

    def _find_stale_audits_due_to_goal(self):
        for goal_id, synced_goal in self.goal_mapping.items():
            filters = {"goal_id": goal_id}
            stale_audits = objects.Audit.list(
                self.ctx, filters=filters, eager=True)

            # Update the goal ID for the stale audits (w/o saving)
            for audit in stale_audits:
                if audit.id not in self.stale_audits_map:
                    audit.goal_id = synced_goal.id
                    self.stale_audits_map[audit.id] = audit
                else:
                    self.stale_audits_map[audit.id].goal_id = synced_goal.id

    def _find_stale_audits_due_to_strategy(self):
        for strategy_id, synced_strategy in self.strategy_mapping.items():
            filters = {"strategy_id": strategy_id}
            stale_audits = objects.Audit.list(
                self.ctx, filters=filters, eager=True)
            # Update strategy IDs for all stale audits (w/o saving)
            for audit in stale_audits:
                if audit.id not in self.stale_audits_map:
                    audit.strategy_id = synced_strategy.id
                    audit.state = objects.audit.State.CANCELLED
                    self.stale_audits_map[audit.id] = audit
                else:
                    self.stale_audits_map[
                        audit.id].strategy_id = synced_strategy.id
                    self.stale_audits_map[
                        audit.id].state = objects.audit.State.CANCELLED

    def _find_stale_action_plans_due_to_strategy(self):
        for strategy_id, synced_strategy in self.strategy_mapping.items():
            filters = {"strategy_id": strategy_id}
            stale_action_plans = objects.ActionPlan.list(
                self.ctx, filters=filters, eager=True)

            # Update strategy IDs for all stale action plans (w/o saving)
            for action_plan in stale_action_plans:
                if action_plan.id not in self.stale_action_plans_map:
                    action_plan.strategy_id = synced_strategy.id
                    action_plan.state = objects.action_plan.State.CANCELLED
                    self.stale_action_plans_map[action_plan.id] = action_plan
                else:
                    self.stale_action_plans_map[
                        action_plan.id].strategy_id = synced_strategy.id
                    self.stale_action_plans_map[
                        action_plan.id].state = (
                            objects.action_plan.State.CANCELLED)

    def _find_stale_action_plans_due_to_audit(self):
        for audit_id, synced_audit in self.stale_audits_map.items():
            filters = {"audit_id": audit_id}
            stale_action_plans = objects.ActionPlan.list(
                self.ctx, filters=filters, eager=True)

            # Update audit IDs for all stale action plans (w/o saving)
            for action_plan in stale_action_plans:
                if action_plan.id not in self.stale_action_plans_map:
                    action_plan.audit_id = synced_audit.id
                    action_plan.state = objects.action_plan.State.CANCELLED
                    self.stale_action_plans_map[action_plan.id] = action_plan
                else:
                    self.stale_action_plans_map[
                        action_plan.id].audit_id = synced_audit.id
                    self.stale_action_plans_map[
                        action_plan.id].state = (
                            objects.action_plan.State.CANCELLED)

    def _soft_delete_removed_goals(self):
        removed_goals = [
            g for g in self.available_goals
            if g.name not in self.discovered_map['goals']]
        for removed_goal in removed_goals:
            removed_goal.soft_delete()
            filters = {"goal_id": removed_goal.id}

            invalid_ats = objects.AuditTemplate.list(self.ctx, filters=filters)
            for at in invalid_ats:
                LOG.warning(
                    "Audit Template '%(audit_template)s' references a "
                    "goal that does not exist", audit_template=at.uuid)

            stale_audits = objects.Audit.list(
                self.ctx, filters=filters, eager=True)
            for audit in stale_audits:
                LOG.warning(
                    "Audit '%(audit)s' references a "
                    "goal that does not exist", audit=audit.uuid)
                if audit.id not in self.stale_audits_map:
                    audit.state = objects.audit.State.CANCELLED
                    self.stale_audits_map[audit.id] = audit
                else:
                    self.stale_audits_map[
                        audit.id].state = objects.audit.State.CANCELLED

    def _soft_delete_removed_strategies(self):
        removed_strategies = [
            s for s in self.available_strategies
            if s.name not in self.discovered_map['strategies']]

        for removed_strategy in removed_strategies:
            removed_strategy.soft_delete()
            filters = {"strategy_id": removed_strategy.id}
            invalid_ats = objects.AuditTemplate.list(self.ctx, filters=filters)
            for at in invalid_ats:
                LOG.info(
                    "Audit Template '%(audit_template)s' references a "
                    "strategy that does not exist",
                    audit_template=at.uuid)
                # In this case we can reset the strategy ID to None
                # so the audit template can still achieve the same goal
                # but with a different strategy
                if at.id not in self.stale_audit_templates_map:
                    at.strategy_id = None
                    self.stale_audit_templates_map[at.id] = at
                else:
                    self.stale_audit_templates_map[at.id].strategy_id = None

            stale_audits = objects.Audit.list(
                self.ctx, filters=filters, eager=True)
            for audit in stale_audits:
                LOG.warning(
                    "Audit '%(audit)s' references a "
                    "strategy that does not exist", audit=audit.uuid)
                if audit.id not in self.stale_audits_map:
                    audit.state = objects.audit.State.CANCELLED
                    self.stale_audits_map[audit.id] = audit
                else:
                    self.stale_audits_map[
                        audit.id].state = objects.audit.State.CANCELLED

            stale_action_plans = objects.ActionPlan.list(
                self.ctx, filters=filters, eager=True)
            for action_plan in stale_action_plans:
                LOG.warning(
                    "Action Plan '%(action_plan)s' references a "
                    "strategy that does not exist",
                    action_plan=action_plan.uuid)
                if action_plan.id not in self.stale_action_plans_map:
                    action_plan.state = objects.action_plan.State.CANCELLED
                    self.stale_action_plans_map[action_plan.id] = action_plan
                else:
                    self.stale_action_plans_map[
                        action_plan.id].state = (
                            objects.action_plan.State.CANCELLED)

    def _soft_delete_removed_scoringengines(self):
        removed_se = [
            se for se in self.available_scoringengines
            if se.name not in self.discovered_map['scoringengines']]
        for se in removed_se:
            LOG.info("Scoring Engine %s removed", se.name)
            se.soft_delete()

    def _discover(self):
        strategies_map = {}
        goals_map = {}
        scoringengines_map = {}
        discovered_map = {
            "goals": goals_map,
            "strategies": strategies_map,
            "scoringengines": scoringengines_map}
        goal_loader = default.DefaultGoalLoader()
        implemented_goals = goal_loader.list_available()

        strategy_loader = default.DefaultStrategyLoader()
        implemented_strategies = strategy_loader.list_available()

        for goal_cls in implemented_goals.values():
            goals_map[goal_cls.get_name()] = GoalMapping(
                name=goal_cls.get_name(),
                display_name=goal_cls.get_translatable_display_name(),
                efficacy_specification=tuple(
                    IndicatorSpec(**indicator.to_dict())
                    for indicator in goal_cls.get_efficacy_specification(
                    ).get_indicators_specifications()))

        for strategy_cls in implemented_strategies.values():
            strategies_map[strategy_cls.get_name()] = StrategyMapping(
                name=strategy_cls.get_name(),
                goal_name=strategy_cls.get_goal_name(),
                display_name=strategy_cls.get_translatable_display_name(),
                parameters_spec=str(strategy_cls.get_schema()))

        for se in scoring_factory.get_scoring_engine_list():
            scoringengines_map[se.get_name()] = ScoringEngineMapping(
                name=se.get_name(),
                description=se.get_description(),
                metainfo=se.get_metainfo())

        return discovered_map

    def _soft_delete_stale_goals(self, goal_map, matching_goals):
        """Soft delete the stale goals

        :param goal_map: discovered goal map
        :type goal_map: :py:class:`~.GoalMapping` instance
        :param matching_goals: list of DB goals matching the goal_map
        :type matching_goals: list of :py:class:`~.objects.Goal` instances
        :returns: A list of soft deleted DB goals (subset of matching goals)
        :rtype: list of :py:class:`~.objects.Goal` instances
        """
        goal_display_name = goal_map.display_name
        goal_name = goal_map.name
        goal_efficacy_spec = goal_map.efficacy_specification

        stale_goals = []
        for matching_goal in matching_goals:
            if (matching_goal.efficacy_specification == goal_efficacy_spec and
                    matching_goal.display_name == goal_display_name):
                LOG.info("Goal %s unchanged", goal_name)
            else:
                LOG.info("Goal %s modified", goal_name)
                matching_goal.soft_delete()
                stale_goals.append(matching_goal)

        return stale_goals

    def _soft_delete_stale_strategies(self, strategy_map, matching_strategies):
        strategy_name = strategy_map.name
        strategy_display_name = strategy_map.display_name
        parameters_spec = strategy_map.parameters_spec

        stale_strategies = []
        for matching_strategy in matching_strategies:
            if (matching_strategy.display_name == strategy_display_name and
                    matching_strategy.goal_id not in self.goal_mapping and
                    matching_strategy.parameters_spec ==
                    ast.literal_eval(parameters_spec)):
                LOG.info("Strategy %s unchanged", strategy_name)
            else:
                LOG.info("Strategy %s modified", strategy_name)
                matching_strategy.soft_delete()
                stale_strategies.append(matching_strategy)

        return stale_strategies

    def _soft_delete_stale_scoringengines(
            self, scoringengine_map, matching_scoringengines):
        se_name = scoringengine_map.name
        se_description = scoringengine_map.description
        se_metainfo = scoringengine_map.metainfo

        stale_scoringengines = []
        for matching_scoringengine in matching_scoringengines:
            if (matching_scoringengine.description == se_description and
                    matching_scoringengine.metainfo == se_metainfo):
                LOG.info("Scoring Engine %s unchanged", se_name)
            else:
                LOG.info("Scoring Engine %s modified", se_name)
                matching_scoringengine.soft_delete()
                stale_scoringengines.append(matching_scoringengine)

        return stale_scoringengines



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 ZTE Corporation
#
# Authors: licanwei <li.canwei2@zte.com.cn>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import random

from oslo_log import log

from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.strategy.strategies import base

LOG = log.getLogger(__name__)


class SavingEnergy(base.SavingEnergyBaseStrategy):
    """Saving Energy Strategy

    *Description*

    Saving Energy Strategy together with VM Workload Consolidation Strategy
    can perform the Dynamic Power Management (DPM) functionality, which tries
    to save power by dynamically consolidating workloads even further during
    periods of low resource utilization. Virtual machines are migrated onto
    fewer hosts and the unneeded  hosts are powered off.

    After consolidation, Saving Energy Strategy produces a solution of powering
    off/on according to the following detailed policy:

    In this policy, a preset number(min_free_hosts_num) is given by user, and
    this min_free_hosts_num describes minimum free compute nodes that users
    expect to have, where "free compute nodes" refers to those nodes unused
    but still powered on.

    If the actual number of unused nodes(in power-on state) is larger than
    the given number, randomly select the redundant nodes and power off them;
    If the actual number of unused nodes(in poweron state) is smaller than
    the given number and there are spare unused nodes(in poweroff state),
    randomly select some nodes(unused,poweroff) and power on them.

    *Requirements*

    In this policy, in order to calculate the min_free_hosts_num,
    users must provide two parameters:

    * One parameter("min_free_hosts_num") is a constant int number.
      This number should be int type and larger than zero.

    * The other parameter("free_used_percent") is a percentage number, which
      describes the quotient of min_free_hosts_num/nodes_with_VMs_num,
      where nodes_with_VMs_num is the number of nodes with VMs running on it.
      This parameter is used to calculate a dynamic min_free_hosts_num.
      The nodes with VMs refer to those nodes with VMs running on it.

    Then choose the larger one as the final min_free_hosts_num.

    *Limitations*

    * at least 2 physical compute hosts

    *Spec URL*

    http://specs.openstack.org/openstack/watcher-specs/specs/pike/implemented/energy-saving-strategy.html
    """

    def __init__(self, config, osc=None):

        super(SavingEnergy, self).__init__(config, osc)
        self._ironic_client = None
        self._nova_client = None

        self.with_vms_node_pool = []
        self.free_poweron_node_pool = []
        self.free_poweroff_node_pool = []
        self.free_used_percent = 0
        self.min_free_hosts_num = 1

    @property
    def ironic_client(self):
        if not self._ironic_client:
            self._ironic_client = self.osc.ironic()
        return self._ironic_client

    @property
    def nova_client(self):
        if not self._nova_client:
            self._nova_client = self.osc.nova()
        return self._nova_client

    @classmethod
    def get_name(cls):
        return "saving_energy"

    @classmethod
    def get_display_name(cls):
        return _("Saving Energy Strategy")

    @classmethod
    def get_translatable_display_name(cls):
        return "Saving Energy Strategy"

    @classmethod
    def get_schema(cls):
        """return a schema of two input parameters

        The standby nodes refer to those nodes unused
        but still poweredon to deal with boom of new instances.
        """
        return {
            "properties": {
                "free_used_percent": {
                    "description": ("a rational number, which describes the"
                                    " quotient of"
                                    " min_free_hosts_num/nodes_with_VMs_num"
                                    " where nodes_with_VMs_num is the number"
                                    " of nodes with VMs"),
                    "type": "number",
                    "default": 10.0
                },
                "min_free_hosts_num": {
                    "description": ("minimum number of hosts without VMs"
                                    " but still powered on"),
                    "type": "number",
                    "default": 1
                },
            },
        }

    def add_action_poweronoff_node(self, node, state):
        """Add an action for node disability into the solution.

        :param node: node
        :param state: node power state, power on or power off
        :return: None
        """
        params = {'state': state,
                  'resource_name': node.hostname}
        self.solution.add_action(
            action_type='change_node_power_state',
            resource_id=node.uuid,
            input_parameters=params)

    def get_hosts_pool(self):
        """Get three pools, with_vms_node_pool, free_poweron_node_pool,

        free_poweroff_node_pool.

        """

        node_list = self.ironic_client.node.list()
        for node in node_list:
            node_info = self.ironic_client.node.get(node.uuid)
            hypervisor_id = node_info.extra.get('compute_node_id', None)
            if hypervisor_id is None:
                LOG.warning(('Cannot find compute_node_id in extra '
                             'of ironic node %s'), node.uuid)
                continue
            hypervisor_node = self.nova_client.hypervisors.get(hypervisor_id)
            if hypervisor_node is None:
                LOG.warning(('Cannot find hypervisor %s'), hypervisor_id)
                continue
            node.hostname = hypervisor_node.hypervisor_hostname
            hypervisor_node = hypervisor_node.to_dict()
            compute_service = hypervisor_node.get('service', None)
            host_name = compute_service.get('host')
            try:
                self.compute_model.get_node_by_name(host_name)
            except exception.ComputeNodeNotFound:
                continue

            if not (hypervisor_node.get('state') == 'up'):
                """filter nodes that are not in 'up' state"""
                continue
            else:
                if (hypervisor_node['running_vms'] == 0):
                    if (node_info.power_state == 'power on'):
                        self.free_poweron_node_pool.append(node)
                    elif (node_info.power_state == 'power off'):
                        self.free_poweroff_node_pool.append(node)
                else:
                    self.with_vms_node_pool.append(node)

    def save_energy(self):

        need_poweron = int(max(
            (len(self.with_vms_node_pool) * self.free_used_percent / 100), (
                self.min_free_hosts_num)))
        len_poweron = len(self.free_poweron_node_pool)
        len_poweroff = len(self.free_poweroff_node_pool)
        if len_poweron > need_poweron:
            for node in random.sample(self.free_poweron_node_pool,
                                      (len_poweron - need_poweron)):
                self.add_action_poweronoff_node(node, 'off')
                LOG.debug("power off %s", node.uuid)
        elif len_poweron < need_poweron:
            diff = need_poweron - len_poweron
            for node in random.sample(self.free_poweroff_node_pool,
                                      min(len_poweroff, diff)):
                self.add_action_poweronoff_node(node, 'on')
                LOG.debug("power on %s", node.uuid)

    def pre_execute(self):
        self._pre_execute()
        self.free_used_percent = self.input_parameters.free_used_percent
        self.min_free_hosts_num = self.input_parameters.min_free_hosts_num

    def do_execute(self, audit=None):
        """Strategy execution phase

        This phase is where you should put the main logic of your strategy.
        """
        self.get_hosts_pool()
        self.save_energy()

    def post_execute(self):
        """Post-execution phase

        This can be used to compute the global efficacy
        """
        self.solution.model = self.compute_model

        LOG.debug(self.compute_model.to_string())



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 b<>com
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from watcher.common import context as wcontext
from watcher.common import exception
from watcher.notifications import audit as audit_notifications
from watcher.notifications import base as notificationbase
from watcher.notifications import exception as exception_notifications
from watcher.notifications import strategy as strategy_notifications
from watcher import objects
from watcher.objects import base
from watcher.objects import fields as wfields

CONF = cfg.CONF


@base.WatcherObjectRegistry.register_notification
class TerseActionPlanPayload(notificationbase.NotificationPayloadBase):
    SCHEMA = {
        'uuid': ('action_plan', 'uuid'),

        'state': ('action_plan', 'state'),
        'global_efficacy': ('action_plan', 'global_efficacy'),

        'created_at': ('action_plan', 'created_at'),
        'updated_at': ('action_plan', 'updated_at'),
        'deleted_at': ('action_plan', 'deleted_at'),
    }

    # Version 1.0: Initial version
    # Version 1.1: Changed 'global_efficacy' type Dictionary to List
    VERSION = '1.1'

    fields = {
        'uuid': wfields.UUIDField(),
        'state': wfields.StringField(),
        'global_efficacy': wfields.FlexibleListOfDictField(nullable=True),
        'audit_uuid': wfields.UUIDField(),
        'strategy_uuid': wfields.UUIDField(nullable=True),

        'created_at': wfields.DateTimeField(nullable=True),
        'updated_at': wfields.DateTimeField(nullable=True),
        'deleted_at': wfields.DateTimeField(nullable=True),
    }

    def __init__(self, action_plan, audit=None, strategy=None, **kwargs):
        super(TerseActionPlanPayload, self).__init__(audit=audit,
                                                     strategy=strategy,
                                                     **kwargs)
        self.populate_schema(action_plan=action_plan)


@base.WatcherObjectRegistry.register_notification
class ActionPlanPayload(TerseActionPlanPayload):
    SCHEMA = {
        'uuid': ('action_plan', 'uuid'),

        'state': ('action_plan', 'state'),
        'global_efficacy': ('action_plan', 'global_efficacy'),

        'created_at': ('action_plan', 'created_at'),
        'updated_at': ('action_plan', 'updated_at'),
        'deleted_at': ('action_plan', 'deleted_at'),
    }

    # Version 1.0: Initial version
    # Vesrsion 1.1: changed global_efficacy type
    VERSION = '1.1'

    fields = {
        'audit': wfields.ObjectField('TerseAuditPayload'),
        'strategy': wfields.ObjectField('StrategyPayload'),
    }

    def __init__(self, action_plan, audit, strategy, **kwargs):
        if not kwargs.get('audit_uuid'):
            kwargs['audit_uuid'] = audit.uuid

        if strategy and not kwargs.get('strategy_uuid'):
            kwargs['strategy_uuid'] = strategy.uuid

        super(ActionPlanPayload, self).__init__(
            action_plan, audit=audit, strategy=strategy, **kwargs)


@base.WatcherObjectRegistry.register_notification
class ActionPlanStateUpdatePayload(notificationbase.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'old_state': wfields.StringField(nullable=True),
        'state': wfields.StringField(nullable=True),
    }


@base.WatcherObjectRegistry.register_notification
class ActionPlanCreatePayload(ActionPlanPayload):
    # Version 1.0: Initial version
    # Version 1.1: Changed global_efficacy_type
    VERSION = '1.1'
    fields = {}

    def __init__(self, action_plan, audit, strategy):
        super(ActionPlanCreatePayload, self).__init__(
            action_plan=action_plan,
            audit=audit,
            strategy=strategy)


@base.WatcherObjectRegistry.register_notification
class ActionPlanUpdatePayload(ActionPlanPayload):
    # Version 1.0: Initial version
    # Version 1.1: Changed global_efficacy_type
    VERSION = '1.1'
    fields = {
        'state_update': wfields.ObjectField('ActionPlanStateUpdatePayload'),
    }

    def __init__(self, action_plan, state_update, audit, strategy):
        super(ActionPlanUpdatePayload, self).__init__(
            action_plan=action_plan,
            state_update=state_update,
            audit=audit,
            strategy=strategy)


@base.WatcherObjectRegistry.register_notification
class ActionPlanActionPayload(ActionPlanPayload):
    # Version 1.0: Initial version
    # Version 1.1: Changed global_efficacy_type
    VERSION = '1.1'
    fields = {
        'fault': wfields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, action_plan, audit, strategy, **kwargs):
        super(ActionPlanActionPayload, self).__init__(
            action_plan=action_plan,
            audit=audit,
            strategy=strategy,
            **kwargs)


@base.WatcherObjectRegistry.register_notification
class ActionPlanDeletePayload(ActionPlanPayload):
    # Version 1.0: Initial version
    # Version 1.1: Changed global_efficacy_type
    VERSION = '1.1'
    fields = {}

    def __init__(self, action_plan, audit, strategy):
        super(ActionPlanDeletePayload, self).__init__(
            action_plan=action_plan,
            audit=audit,
            strategy=strategy)


@base.WatcherObjectRegistry.register_notification
class ActionPlanCancelPayload(ActionPlanPayload):
    # Version 1.0: Initial version
    # Version 1.1: Changed global_efficacy_type
    VERSION = '1.1'
    fields = {
        'fault': wfields.ObjectField('ExceptionPayload', nullable=True),
    }

    def __init__(self, action_plan, audit, strategy, **kwargs):
        super(ActionPlanCancelPayload, self).__init__(
            action_plan=action_plan,
            audit=audit,
            strategy=strategy,
            **kwargs)


@notificationbase.notification_sample('action_plan-execution-error.json')
@notificationbase.notification_sample('action_plan-execution-end.json')
@notificationbase.notification_sample('action_plan-execution-start.json')
@base.WatcherObjectRegistry.register_notification
class ActionPlanActionNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionPlanActionPayload')
    }


@notificationbase.notification_sample('action_plan-create.json')
@base.WatcherObjectRegistry.register_notification
class ActionPlanCreateNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionPlanCreatePayload')
    }


@notificationbase.notification_sample('action_plan-update.json')
@base.WatcherObjectRegistry.register_notification
class ActionPlanUpdateNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionPlanUpdatePayload')
    }


@notificationbase.notification_sample('action_plan-delete.json')
@base.WatcherObjectRegistry.register_notification
class ActionPlanDeleteNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionPlanDeletePayload')
    }


@notificationbase.notification_sample('action_plan-cancel-error.json')
@notificationbase.notification_sample('action_plan-cancel-end.json')
@notificationbase.notification_sample('action_plan-cancel-start.json')
@base.WatcherObjectRegistry.register_notification
class ActionPlanCancelNotification(notificationbase.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': wfields.ObjectField('ActionPlanCancelPayload')
    }


def _get_common_payload(action_plan):
    audit = None
    strategy = None
    try:
        audit = action_plan.audit
        strategy = action_plan.strategy
    except NotImplementedError:
        raise exception.EagerlyLoadedActionPlanRequired(
            action_plan=action_plan.uuid)

    goal = objects.Goal.get(
        wcontext.make_context(show_deleted=True), audit.goal_id)
    audit_payload = audit_notifications.TerseAuditPayload(
        audit=audit, goal_uuid=goal.uuid)

    strategy_payload = strategy_notifications.StrategyPayload(
        strategy=strategy)

    return audit_payload, strategy_payload


def send_create(context, action_plan, service='infra-optim', host=None):
    """Emit an action_plan.create notification."""
    audit_payload, strategy_payload = _get_common_payload(action_plan)

    versioned_payload = ActionPlanCreatePayload(
        action_plan=action_plan,
        audit=audit_payload,
        strategy=strategy_payload,
    )

    notification = ActionPlanCreateNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action_plan',
            action=wfields.NotificationAction.CREATE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_update(context, action_plan, service='infra-optim',
                host=None, old_state=None):
    """Emit an action_plan.update notification."""
    audit_payload, strategy_payload = _get_common_payload(action_plan)

    state_update = ActionPlanStateUpdatePayload(
        old_state=old_state,
        state=action_plan.state if old_state else None)

    versioned_payload = ActionPlanUpdatePayload(
        action_plan=action_plan,
        state_update=state_update,
        audit=audit_payload,
        strategy=strategy_payload,
    )

    notification = ActionPlanUpdateNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action_plan',
            action=wfields.NotificationAction.UPDATE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_delete(context, action_plan, service='infra-optim', host=None):
    """Emit an action_plan.delete notification."""
    audit_payload, strategy_payload = _get_common_payload(action_plan)

    versioned_payload = ActionPlanDeletePayload(
        action_plan=action_plan,
        audit=audit_payload,
        strategy=strategy_payload,
    )

    notification = ActionPlanDeleteNotification(
        priority=wfields.NotificationPriority.INFO,
        event_type=notificationbase.EventType(
            object='action_plan',
            action=wfields.NotificationAction.DELETE),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_action_notification(context, action_plan, action, phase=None,
                             priority=wfields.NotificationPriority.INFO,
                             service='infra-optim', host=None):
    """Emit an action_plan action notification."""
    audit_payload, strategy_payload = _get_common_payload(action_plan)

    fault = None
    if phase == wfields.NotificationPhase.ERROR:
        fault = exception_notifications.ExceptionPayload.from_exception()

    versioned_payload = ActionPlanActionPayload(
        action_plan=action_plan,
        audit=audit_payload,
        strategy=strategy_payload,
        fault=fault,
    )

    notification = ActionPlanActionNotification(
        priority=priority,
        event_type=notificationbase.EventType(
            object='action_plan',
            action=action,
            phase=phase),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)


def send_cancel_notification(context, action_plan, action, phase=None,
                             priority=wfields.NotificationPriority.INFO,
                             service='infra-optim', host=None):
    """Emit an action_plan cancel notification."""
    audit_payload, strategy_payload = _get_common_payload(action_plan)

    fault = None
    if phase == wfields.NotificationPhase.ERROR:
        fault = exception_notifications.ExceptionPayload.from_exception()

    versioned_payload = ActionPlanCancelPayload(
        action_plan=action_plan,
        audit=audit_payload,
        strategy=strategy_payload,
        fault=fault,
    )

    notification = ActionPlanCancelNotification(
        priority=priority,
        event_type=notificationbase.EventType(
            object='action_plan',
            action=action,
            phase=phase),
        publisher=notificationbase.NotificationPublisher(
            host=host or CONF.host,
            binary=service),
        payload=versioned_payload)

    notification.emit(context)
